/*----------------------------------------------------------------------------
	This is a Virtual Desktop Manager for use under Win32 Shells that
	support the LiteStep 0.24.5 module standard.

	Copyright (C) 2001-2003 Chris Rempel

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
----------------------------------------------------------------------------*/

/*---System Includes--------------------------------------------------------*/
#include <tchar.h>
#include <windows.h>

/*---Module Includes--------------------------------------------------------*/
#include "common.h"

/*--------------------------------------------------------------------------*/
/* If you have an old Platform SDK, then uncomment the following line.      */
//#define GET_SWITCHTOTHISWINDOW
#ifdef GET_SWITCHTOTHISWINDOW
FUNC_VOID__HWND_BOOL SwitchToThisWindow;
#endif

/*
 * Half way through 0.32 I changed the way I enumerate the windows.  So, I am
 * keeping the old way around for a bit, until I'm satisfied the new way is
 * as solid as the old.  In actuallity the new way should be better, because
 * I no longer call GetWindow() in a loop which according to MSDN may cause
 * problems.
 */
//#define USE_OLDWINDOWENUMERATION

typedef struct _STICKYITEM
{
	_TCHAR szClass[256];
	_TCHAR szTitle[256];
	struct _STICKYITEM *pNext;
} STICKYITEM, *PSTICKYITEM;


/*---Global Variables-------------------------------------------------------*/
#ifndef USE_OLDWINDOWENUMERATION
HDWP g_hDwp;
HWND g_hIgnoreActive;
#else
HWND g_hTopWnd;
#endif /* USE_OLDWINDOWENUMERATION */
UINT g_unCurDesk, g_unDesks;
RECT g_DTRect, g_WARect;
INT g_nDTWidth, g_nDTHeight;
INT g_nWAWidth, g_nWAHeight;
BOOL g_bFocusLast, g_bKeepMaximized, g_bVertical;
BOOL g_bAutoGather, g_bStrictGather;
LPTSTR* g_pCmdList;
LPTSTR g_pszQuitCmd;
PSTICKYITEM g_pStickyList;
/*--------------------------------------------------------------------------*/


/*--------executeCmd----------------------------------------------------------
 Purpose:
 - To execute the command associated with the specified Virtual Desktop when
   switching to that VD.
----------------------------------------------------------------------------*/
VOID executeCmd( UINT const unVD )
{
	/*------------------------------------------------------------------------
	 If the specified VD Is valid, and we have populated a command list, and
	 a command was defined for the specified desktop, then execute it.
	------------------------------------------------------------------------*/
	if(VD_ISVALID(unVD) && NULL != g_pCmdList && NULL != g_pCmdList[unVD])
	{
		LSExecute(NULL, g_pCmdList[unVD], 0);
	}

	return;
}

/*--------isSticky------------------------------------------------------------
 Purpose:
 - To determine whether the specified window has been defined by the user to
   stay on the current desktop at all times.
----------------------------------------------------------------------------*/
BOOL isSticky( HWND const hWnd )
{
	PSTICKYITEM pItem;
	_TCHAR text[256];
	BOOL bRet;

	bRet = FALSE;

	/*------------------------------------------------------------------------
	 Itterate through the sticky list until it is found, or we reach the end
	 of the list.
	------------------------------------------------------------------------*/
	for(pItem = g_pStickyList; !bRet && NULL != pItem; pItem = pItem->pNext)
	{
		/*--------------------------------------------------------------------
		 Check to see if we were able to get the Class Name.  If so, match it
		 against the user defined class name pattern.  If it matches, get the
		 Window Title and if successful, match it against the user defined
		 window title pattern.  If it matched both Class and Title, then flag
		 the window as Sticky.

		 NOTE: GetClassName always NULL terminates it's buffer. So no worries.
		--------------------------------------------------------------------*/
		if(GetClassName(hWnd, text, 256))
		{
			if(match(pItem->szClass, text))
			{
				text[0] = 0;

				if(SendMessage(hWnd, WM_GETTEXT, (WPARAM)256, (LPARAM)text))
				{
					/* Make sure our buffer is NULL terminated. */
					text[255] = 0;
				}

				if(match(pItem->szTitle, text))
				{
					bRet = TRUE;
				}
			}
		}
	}

	return(bRet);
}

/*--------ignoreWindow--------------------------------------------------------
 Purpose:
 - To determine whether the specified window is to be ignored by our code.
   Basically, this is a base check.  We do not care about minimized windows or
   hidden windows (duh!) or Shell Windows.
----------------------------------------------------------------------------*/
BOOL ignoreWindow( HWND const hWnd )
{
	//LONG lStyle = GetWindowLong(hWnd, GWL_STYLE);

	/*------------------------------------------------------------------------
	 Check to see if it is minimized, or not visible, or a shell window.
	------------------------------------------------------------------------*/
	return(
		IsIconic(hWnd)
		||
		!IsWindowVisible(hWnd)
		//WS_MINIMIZE == (lStyle & WS_MINIMIZE)
		//||
		//WS_VISIBLE != (lStyle & WS_VISIBLE)
		||
		SHELL_WINDOW_ID == GetWindowLong(hWnd, GWL_USERDATA)
	);
}

/*--------isOmniWindow--------------------------------------------------------
 Purpose:
 - To determine whether the specified window is to stay on the current desktop
   at all times (or in other words, we simply do not modify it's position).
   This happens if it is minimized, not visible, a shell window, or the user
   specified the window as "Sticky".
----------------------------------------------------------------------------*/
BOOL isOmniWindow( HWND const hWnd )
{
	return(ignoreWindow(hWnd) || isSticky(hWnd));
}

/*--------isWindowInVD--------------------------------------------------------
 Purpose:
 - To determine whether the specified window is located on the specified
   Virtual Desktop.

 NOTE: We want to make this more intelligent than just taking the top-left
 point of the window and seeing if it is on the desktop.  A better solution
 would be to see if the majority of the window is on the specified desktop.
 The only problem with this, is that some users like to drag a window off the
 side of the screen temporarily while working, and this could lead to that
 window having the majority of its rectangle on another desktop.  I'm sure
 there is a good solution to this.  Yah, be optimistic.  :P

 NOTE: Obviously this function isn't complete as it ignores unVD param. :)
----------------------------------------------------------------------------*/
BOOL isWindowInVD( HWND const hWnd, UINT const unVD )
{
	RECT rWnd;

	/*------------------------------------------------------------------------
	 Get the specified window's rectangle.
	------------------------------------------------------------------------*/
	GetWindowRect(hWnd, &rWnd);

	/*------------------------------------------------------------------------
	 A simple check to see if it is a valid rectangle.
	------------------------------------------------------------------------*/
	if(!IsRectEmpty(&rWnd))
	{
		//POINT pt;

		/*--------------------------------------------------------------------
		 Create a POINT from the window rectangle's top-left point.
		--------------------------------------------------------------------*/
		//pt.x = rWnd.left;
		//pt.y = rWnd.top;

		/*--------------------------------------------------------------------
		 Return whether the point (and thus the window) is contained within
		 the DeskTop rectangle.
		--------------------------------------------------------------------*/
		//return(PtInRect(&g_DTRect, pt));

		return(
			g_bVertical
			?
			!(rWnd.top > g_nDTHeight || rWnd.top < -(VD_SPACING))
			:
			!(rWnd.left > g_nDTWidth || rWnd.left < -(VD_SPACING))
		);
	}

	/*------------------------------------------------------------------------
	 Default fail case.
	------------------------------------------------------------------------*/
	return(FALSE);
}

/*--------isWindowInCurrentVD-------------------------------------------------
 Purpose:
 - To determine whether the specified window is located on the current Virtual
   Desktop.
----------------------------------------------------------------------------*/
BOOL isWindowInCurrentVD( HWND const hWnd )
{
	/*------------------------------------------------------------------------
	 Simply make this function a wrapper to isWindowInVD and pass in the
	 current desktop number.
	------------------------------------------------------------------------*/
	return(isWindowInVD(hWnd, g_unCurDesk));
}

/*--------------------------------------------------------------------------*/
VOID updateRectH( RECT * const r )
{
	INT size;

	size = abs(r->right - r->left);

	/* Initial x/width position calculation */
	if(r->left > g_nDTWidth)
	{
		r->left %= VD_WIDTH;
		if(r->left > g_nDTWidth) r->left -= VD_WIDTH;
		r->right = r->left + size;
	}
	else if(r->left < -(VD_SPACING))
	{
		r->left %= VD_WIDTH;
		if(r->left < -(VD_SPACING)) r->left += VD_WIDTH;
		r->right = r->left + size;
	}

	/* Don't allow a window to be in more then one VD at a time if possible */
	if(!g_bStrictGather && r->right > VD_WIDTH)
	{
		if(r->left >= (r->right - g_nDTWidth))
		{
			r->left -= (r->right - g_nDTWidth);
		}
		else if((size - g_nDTWidth) < VD_SPACING)
		{
			r->left = (size - g_nDTWidth) / (-2);
		}
		else
		{
			r->left = 0;
		}

		r->right = r->left + size;
	}

	return;
}

/*--------------------------------------------------------------------------*/
VOID updateRectV( RECT * const r )
{
	INT size;

	size = abs(r->bottom - r->top);

	/* Initial x/width position calculation */
	if(r->top > g_nDTHeight)
	{
		r->top %= VD_HEIGHT;
		if(r->top > g_nDTHeight) r->top -= VD_HEIGHT;
		r->bottom = r->top + size;
	}
	else if(r->top < -(VD_SPACING))
	{
		r->top %= VD_HEIGHT;
		if(r->top < -(VD_SPACING)) r->top += VD_HEIGHT;
		r->bottom = r->top + size;
	}

	/* Don't allow a window to be in more then one VD at a time if possible */
	if(!g_bStrictGather && r->bottom > VD_HEIGHT)
	{
		if(r->top >= (r->bottom - g_nDTHeight))
		{
			r->top -= (r->bottom - g_nDTHeight);
		}
		else if((size - g_nDTHeight) < VD_SPACING)
		{
			r->top = (size - g_nDTHeight) / (-2);
		}
		else
		{
			r->top = 0;
		}

		r->bottom = r->top + size;
	}

	return;
}

#ifndef USE_OLDWINDOWENUMERATION
/*--------------------------------------------------------------------------*/
BOOL CALLBACK EnumProcSwitchTo( HWND hWnd, LPARAM lParam )
{
	if( !isOmniWindow(hWnd) && g_hIgnoreActive != hWnd )
	{
		RECT r;

		if( !g_bKeepMaximized && IsZoomed(hWnd) )
		{
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);
		}

		GetWindowRect(hWnd, &r);
		if( g_bVertical )
		{
			r.top -= (INT)lParam;
		}
		else
		{
			r.left -= (INT)lParam;
		}

		g_hDwp = DeferWindowPos(
			 g_hDwp
			,hWnd
			,NULL
			,r.left
			,r.top
			,0
			,0
			,SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE
		);
	}

	return(TRUE);
}

/*--------------------------------------------------------------------------*/
BOOL CALLBACK EnumProcGatherTo( HWND hWnd, LPARAM lParam )
{
	if( !isOmniWindow(hWnd) )
	{
		RECT r;
		char const cType = (char const)lParam;

		if( !g_bKeepMaximized && IsZoomed(hWnd) )
		{
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);
		}

		GetWindowRect(hWnd, &r);

		if( 0x00 != (cType & VD_HORIZONTAL) )
		{
			updateRectH(&r);
		}
		if( 0x00 != (cType & VD_VERTICAL) )
		{
			updateRectV(&r);
		}

		g_hDwp = DeferWindowPos(
			 g_hDwp
			,hWnd
			,NULL
			,r.left
			,r.top
			,0
			,0
			,SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE
		);
	}

	return(TRUE);
}

/*--------------------------------------------------------------------------*/
BOOL CALLBACK EnumProcFocusLast( HWND hWnd, LPARAM lParam )
{
	if(WS_EX_TOPMOST != (GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_TOPMOST))
	{
		if(!ignoreWindow(hWnd) && isWindowInCurrentVD(hWnd))
		{
			SwitchToThisWindow(GetLastActivePopup(hWnd), TRUE);
			return(FALSE);
		}
	}

	return(TRUE);
	UNREFERENCED_PARAMETER( lParam );
}
#else /* USE_OLDWINDOWENUMERATION */
/*--------------------------------------------------------------------------*/
BOOL CALLBACK EnumProcGetFirstWindow( HWND hWnd, LPARAM lParam )
{
	if(WS_EX_TOPMOST != (GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_TOPMOST))
	{
		g_hTopWnd = hWnd;
		return(FALSE);
	}

	return(TRUE);
	UNREFERENCED_PARAMETER( lParam );
}

/*--------------------------------------------------------------------------*/
BOOL CALLBACK EnumProcGetFirstWindowInVD( HWND hWnd, LPARAM lParam )
{
	if(WS_EX_TOPMOST != (GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_TOPMOST))
	{
		if(!ignoreWindow(hWnd) && isWindowInCurrentVD(hWnd))
		{
			g_hTopWnd = hWnd;
			return(FALSE);
		}
	}

	return(TRUE);
	UNREFERENCED_PARAMETER( lParam );
}
#endif /* USE_OLDWINDOWENUMERATION */

#ifndef USE_OLDWINDOWENUMERATION
/*--------------------------------------------------------------------------*/
VOID focusLast( VOID )
{
	if(g_bFocusLast)
	{
		EnumWindows(EnumProcFocusLast, 0);
	}

	return;
}
#else /* USE_OLDWINDOWENUMERATION */
/*--------------------------------------------------------------------------*/
VOID focusLast( VOID )
{
	if(g_bFocusLast)
	{
		//HWND hWnd;

		g_hTopWnd = NULL;
		EnumWindows(EnumProcGetFirstWindowInVD, 0);

		//hWnd = g_hTopWnd;

		//while(NULL != hWnd)
		//{
		//	if(!ignoreWindow(hWnd) && isWindowInCurrentVD(hWnd))
		//	{
		//		g_hTopWnd = hWnd;
		//	}
		//	hWnd = GetWindow(hWnd, GW_HWNDPREV);
		//}

		SwitchToThisWindow(GetLastActivePopup(g_hTopWnd), TRUE);
	}

	return;
}
#endif /* USE_OLDWINDOWENUMERATION */

#ifndef USE_OLDWINDOWENUMERATION
/*--------------------------------------------------------------------------*/
VOID switchToVD( UINT const unVD )
{
	if( unVD != g_unCurDesk && VD_ISVALID(unVD) )
	{
		INT nAdd;

		nAdd = (INT)(
			((unVD % g_unDesks) - (g_unCurDesk % g_unDesks))
			*
			(g_bVertical ? VD_HEIGHT:VD_WIDTH)
		);

		g_hDwp = BeginDeferWindowPos(1);

		EnumWindows(EnumProcSwitchTo, (LPARAM)nAdd);

		EndDeferWindowPos(g_hDwp);

		g_unCurDesk = unVD;

		focusLast();

		executeCmd(g_unCurDesk);
	}

	return;
}
#else /* USE_OLDWINDOWENUMERATION */
/*--------------------------------------------------------------------------*/
VOID switchToVD( UINT const unVD )
{
	if( unVD != g_unCurDesk && VD_ISVALID(unVD) )
	{
		INT add;
		HWND hWnd;
		HDWP dwp;
		RECT r;

		add = (INT)(
			((unVD % g_unDesks) - (g_unCurDesk % g_unDesks))
			*
			(g_bVertical ? VD_HEIGHT:VD_WIDTH)
		);

		hWnd = NULL;
		g_hTopWnd = NULL;
		EnumWindows(EnumProcGetFirstWindow, 0);

		if( NULL != g_hTopWnd )
		{
			hWnd = GetWindow(g_hTopWnd, GW_HWNDLAST);
		}

		dwp = BeginDeferWindowPos(1);

		while( NULL != hWnd )
		{
			if( !isOmniWindow(hWnd) )
			{
				if( !g_bKeepMaximized && IsZoomed(hWnd) )
				{
					ShowWindow(hWnd, SW_SHOWNOACTIVATE);
				}

				GetWindowRect(hWnd, &r);
				if( g_bVertical )
				{
					r.top    -= add;
					//r.bottom -= add;
				}
				else
				{
					r.left  -= add;
					//r.right -= add;
				}

				dwp = DeferWindowPos(
					 dwp
					,hWnd
					,NULL
					,r.left
					,r.top
					,0 //r.right - r.left
					,0 //r.bottom - r.top
					,SWP_NOZORDER|SWP_NOACTIVATE/*Using NOSIZE*/|SWP_NOSIZE
				);
			}

			hWnd = GetWindow(hWnd, GW_HWNDPREV);
		}

		EndDeferWindowPos(dwp);

		g_unCurDesk = unVD;

		focusLast();

		executeCmd(g_unCurDesk);
	}

	return;
}
#endif /* USE_OLDWINDOWENUMERATION */

#ifndef USE_OLDWINDOWENUMERATION
/*--------------------------------------------------------------------------*/
VOID gatherToVD( UINT const unVD, char const cType )
{
	if( VD_ISVALID(unVD) )
	{
		UINT unOldDesk;

		g_hDwp = BeginDeferWindowPos(1);

		EnumWindows(EnumProcGatherTo, (LPARAM)cType);

		EndDeferWindowPos(g_hDwp);

		unOldDesk = g_unCurDesk;
		g_unCurDesk = unVD;

		focusLast();

		if( unVD != unOldDesk )
		{
			executeCmd(g_unCurDesk);
		}
	}

	return;
}
#else /* USE_OLDWINDOWENUMERATION */
/*--------------------------------------------------------------------------*/
VOID gatherToVD( UINT const unVD, char const cType )
{
	if( VD_ISVALID(unVD) )
	{
		//UINT unFlags;
		UINT unOldDesk;
		HWND hWnd;
		HDWP dwp;
		RECT r;

		hWnd = NULL;
		g_hTopWnd = NULL;
		EnumWindows(EnumProcGetFirstWindow, 0);

		if( NULL != g_hTopWnd )
		{
			hWnd = GetWindow(g_hTopWnd, GW_HWNDLAST);
		}

		dwp = BeginDeferWindowPos(1);

		while( NULL != hWnd )
		{
			if( !isOmniWindow(hWnd) )
			{
				if( !g_bKeepMaximized && IsZoomed(hWnd) )
				{
					ShowWindow(hWnd, SW_SHOWNOACTIVATE);
				}

				GetWindowRect(hWnd, &r);

				if( 0x00 != (cType & VD_HORIZONTAL) )
				{
					updateRectH(&r);
				}
				if( 0x00 != (cType & VD_VERTICAL) )
				{
					updateRectV(&r);
				}

				/*
				unFlags = SWP_NOZORDER|SWP_NOACTIVATE|(
					g_bStrictGather || (g_bKeepMaximized && IsZoomed(hWnd))
					?
					SWP_NOSIZE : 0
				);
				*/

				dwp = DeferWindowPos(
					 dwp
					,hWnd
					,NULL
					,r.left
					,r.top
					,r.right - r.left
					,r.bottom - r.top
					,SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE //unFlags
				);
			}
		}

		EndDeferWindowPos(dwp);

		unOldDesk = g_unCurDesk;
		g_unCurDesk = unVD;

		focusLast();

		if( unVD != unOldDesk )
		{
			executeCmd(g_unCurDesk);
		}
	}

	return;
}
#endif /* USE_OLDWINDOWENUMERATION */

/*--------------------------------------------------------------------------*/
VOID moveWndTo( HWND const hWnd, UINT const unVD )
{
#ifndef USE_OLDWINDOWENUMERATION
	if( VD_ISVALID(unVD) )
	{
		UINT unOldDesk;

		g_hIgnoreActive = hWnd;

		g_hDwp = BeginDeferWindowPos(1);

		if( unVD != g_unCurDesk )
		{
			INT nAdd;

			nAdd = (INT)(
				((unVD % g_unDesks) - (g_unCurDesk % g_unDesks))
				*
				(g_bVertical ? VD_HEIGHT:VD_WIDTH)
			);

			EnumWindows(EnumProcSwitchTo, (LPARAM)nAdd);
		}

		EnumProcGatherTo(hWnd, (LPARAM)(char const)VD_HORZVERT);

		EndDeferWindowPos(g_hDwp);

		g_hIgnoreActive = NULL;

		unOldDesk = g_unCurDesk;
		g_unCurDesk = unVD;

		if( unVD != unOldDesk )
		{
			executeCmd(g_unCurDesk);
		}
	}
#endif /* USE_OLDWINDOWENUMERATION */
	return;
}

/*--------------------------------------------------------------------------*/
VOID moveActiveWndTo( UINT const unVD )
{
	moveWndTo(GetLastActivePopup(GetForegroundWindow()), unVD);

	return;
}

/*--------------------------------------------------------------------------*/
VOID stretchWnd( HWND const hWnd, enum EbvdmStretchMode const enStretch )
{
	RECT rWnd;

	GetWindowRect(hWnd, &rWnd);

	switch( enStretch )
	{
	case eBVDM_STRETCH__VERTICAL:
		SetWindowPos(
			 hWnd
			,NULL
			,rWnd.left
			,g_WARect.top
			,rWnd.right - rWnd.left
			,g_nWAHeight
			,SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER
		);
		break;

	case eBVDM_STRETCH__HORIZONTAL:
		SetWindowPos(
			 hWnd
			,NULL
			,g_WARect.left
			,rWnd.top
			,g_nWAWidth
			,rWnd.bottom - rWnd.top
			,SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER
		);
		break;

	case eBVDM_STRETCH__BOTH:
		SetWindowPos(
			 hWnd
			,NULL
			,g_WARect.left
			,g_WARect.top
			,g_nWAWidth
			,g_nWAHeight
			,SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER
		);
		break;
	}

	return;
}

/*--------------------------------------------------------------------------*/
VOID alignWnd( HWND const hWnd, enum EbvdmAlignMode const enAlign )
{
	RECT rWnd;
	INT nX, nY;

	GetWindowRect(hWnd, &rWnd);

	switch( enAlign )
	{
	case eBVDM_ALIGN__LEFT:
		nX = g_WARect.left;
		nY = rWnd.top;
		break;

	case eBVDM_ALIGN__RIGHT:
		nX = g_WARect.right - (rWnd.right - rWnd.left);
		nY = rWnd.top;
		break;

	case eBVDM_ALIGN__TOP:
		nX = rWnd.left;
		nY = g_WARect.top;
		break;

	case eBVDM_ALIGN__BOTTOM:
		nX = rWnd.left;
		nY = g_WARect.bottom - (rWnd.bottom - rWnd.top);
		break;

	case eBVDM_ALIGN__CENTER:
		nX = g_WARect.top + ((g_nWAWidth - (rWnd.right - rWnd.left)) / 2);
		nY = rWnd.top;
		break;

	case eBVDM_ALIGN__MIDDLE:
		nX = rWnd.left;
		nY = g_WARect.left + ((g_nWAHeight - (rWnd.bottom - rWnd.top)) / 2);
		break;

	default:
		return;
	}

	SetWindowPos(
		 hWnd
		,NULL
		,nX
		,nY
		,0
		,0
		,SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER|SWP_NOSIZE
	);

	return;
}

/*--------------------------------------------------------------------------*/
VOID readQuitCmd( VOID )
{
	_TCHAR buffer[4096];

	*buffer = 0;
	GetRCLine( _T("bVWMQuitCmd"), buffer, 4096, _T("") );

	if(0 != *buffer)
	{
		g_pszQuitCmd = (LPTSTR)malloc((_tcslen(buffer) + 1) * sizeof(_TCHAR));

		if(NULL != g_pszQuitCmd)
		{
			_tcscpy(g_pszQuitCmd, buffer);
		}
	}

	return;
}

/*--------------------------------------------------------------------------*/
VOID freeQuitCmd( VOID )
{
	if(NULL != g_pszQuitCmd)
	{
		free(g_pszQuitCmd);
		g_pszQuitCmd = NULL;
	}

	return;
}

/*--------------------------------------------------------------------------*/
VOID freeCommands( VOID )
{
	if(NULL != g_pCmdList)
	{
		UINT i;

		for(i = 0; i < g_unDesks; i++)
		{
			if(NULL != g_pCmdList[i])
			{
				free(g_pCmdList[i]);
			}
		}

		free(g_pCmdList);
		g_pCmdList = NULL;
	}

	return;
}

/*--------------------------------------------------------------------------*/
VOID readCommands( VOID )
{
	LPTSTR tokens[2];
	_TCHAR buffer[4096];
	_TCHAR token[128];
	_TCHAR cmd[4096];
	ULONG uLCHandle;
	INT nCount;

	uLCHandle = LCOpen(NULL);

	if(0 == uLCHandle)
	{
		return;
	}

	freeCommands();

	tokens[0] = token;
	tokens[1] = token;

	while(LCReadNextConfig(uLCHandle, _T("*bVWMExec"), buffer, sizeof(buffer)/sizeof(_TCHAR)))
	{
		*cmd = 0;
		nCount = LCTokenize(buffer, tokens, 2, cmd);

		if(nCount >= 2 && 0 != *cmd)
		{
			nCount = _ttoi(tokens[1]);

			if(nCount < 0 || (UINT)nCount >= g_unDesks)
			{
				continue;
			}

			if(NULL == g_pCmdList)
			{
				g_pCmdList = (LPTSTR*)calloc(g_unDesks, sizeof(LPTSTR));
			}

			if(NULL != g_pCmdList && NULL == g_pCmdList[nCount])
			{
				g_pCmdList[nCount] = (LPTSTR)malloc(
					(_tcslen(cmd) + 1) * sizeof(_TCHAR)
				);

				if(NULL != g_pCmdList[nCount])
				{
					_tcscpy(g_pCmdList[nCount], cmd);
				}
			}
		}
	}

	LCClose(uLCHandle);

	return;
}

/*--------------------------------------------------------------------------*/
VOID freeSticky( VOID )
{
	PSTICKYITEM pNext;

	while(NULL != g_pStickyList)
	{
		pNext = g_pStickyList->pNext;
		free(g_pStickyList);
		g_pStickyList = pNext;
	}

	return;
}

/*--------------------------------------------------------------------------*/
VOID readSticky( VOID )
{
	PSTICKYITEM pItem, pLast;
	LPTSTR tokens[3];
	_TCHAR buffer[4096];
	_TCHAR szTitle[256];
	_TCHAR szClass[256];
	ULONG uLCHandle;
	INT nCount;

	uLCHandle = LCOpen(NULL);

	if(0 == uLCHandle)
	{
		return;
	}

	freeSticky();
	/* Not necessary, but gets rid of compiler warning */
	pLast = NULL;

	tokens[0] = szClass;
	tokens[1] = szClass;
	tokens[2] = szTitle;

	while(LCReadNextConfig(uLCHandle, _T("*bVWMSticky"), buffer, sizeof(buffer)/sizeof(_TCHAR)))
	{
		nCount = LCTokenize(buffer, tokens, 3, NULL);

		if(nCount >= 2)
		{
			pItem = malloc(sizeof(STICKYITEM));
			memset(pItem, 0, sizeof(STICKYITEM));

			_tcsncpy(pItem->szClass, szClass, 256);
			pItem->szClass[255] = 0;

			if(nCount > 2)
			{
				_tcsncpy(pItem->szTitle, szTitle, 256);
				pItem->szTitle[255] = 0;
			}
			else
			{
				_tcscpy(pItem->szTitle, _T("*"));
			}

			if(NULL == g_pStickyList)
			{
				pLast = g_pStickyList = pItem;
			}
			else
			{
				while(NULL != pLast->pNext)
				{
					pLast = pLast->pNext;
				}

				pLast = pLast->pNext = pItem;
			}
		}
	}

	LCClose(uLCHandle);

	return;
}

/*--------------------------------------------------------------------------*/
VOID setNumDesks( UINT const unNew )
{
	g_unDesks = (2 > unNew) ? 2 : unNew;

	freeCommands();
	readCommands();

	// Call gather command for windows outside of new space

	return;
}

/*--------------------------------------------------------------------------*/
VOID loadSettings( VOID )
{
	RECT r;

	//g_nDTWidth = GetSystemMetrics(SM_CXSCREEN);
	//g_nDTHeight = GetSystemMetrics(SM_CYSCREEN);

	GetWindowRect(GetDesktopWindow(), &r);

	g_nDTWidth = g_nWAWidth = r.right - r.left;
	g_nDTHeight = g_nWAHeight = r.bottom - r.top;

	CopyRect(&g_DTRect, &r);
	CopyRect(&g_WARect, &r);

	if( SystemParametersInfo(SPI_GETWORKAREA, 0, &r, 0) )
	{
		g_nWAWidth = r.right - r.left;
		g_nWAHeight = r.bottom - r.top;

		CopyRect(&g_WARect, &r);
	}

	g_bFocusLast = (BOOL)GetRCBool(_T("bVWMFocusLast"), TRUE);
	g_bKeepMaximized = (BOOL)GetRCBool(_T("bVWMKeepMaximized"), TRUE);
	g_bAutoGather = (BOOL)GetRCBool(_T("bVWMAutoGather"), TRUE);
	g_bVertical = GetRCBool(_T("bVWMVertical"), TRUE);
	g_bStrictGather = GetRCBool(_T("bVWMStrictGather"), TRUE);

	setNumDesks((UINT)GetRCInt(_T("bVWMDesks"), 2));

	//readCommands(); <- done in setNumDesks()
	readSticky();
	readQuitCmd();

#ifdef GET_SWITCHTOTHISWINDOW
	SwitchToThisWindow = (FUNC_VOID__HWND_BOOL)GetProcAddress(
		 GetModuleHandle(_T("USER32"))
		,"SwitchToThisWindow"
	);

	if( NULL == SwitchToThisWindow )
	{
		g_bFocusLast = FALSE;
	}
#endif /* GET_SWITCHTOTHISWINDOW */

	return;
}

/*--------------------------------------------------------------------------*/
VOID unloadSettings( VOID )
{
	freeQuitCmd();
	freeSticky();
	freeCommands();

	return;
}

/*--------------------------------------------------------------------------*/
VOID executeQuitCmd( VOID )
{
	if(NULL != g_pszQuitCmd)
	{
		LSExecute(NULL, g_pszQuitCmd, 0);
	}
	else /* Default quit command */
	{
		switchToVD(0);
	}

	return;
}

/*--------------------------------------------------------------------------*/
